@@ -1008,21 +1008,36 @@ class RpcKombuHATestCase(test_utils.BaseTestCase):
         self.assertRaises(oslo_messaging.MessageDeliveryFailure,
                           self.connection.ensure, mock_callback,
                           retry=4)
-        self.assertEqual(6, mock_callback.call_count)
+        # TODO(stephenfin): Remove when we drop support for kombu < 5.2.4
+        expected = 5
+        if kombu.VERSION < (5, 2, 4):
+            expected = 6
+        self.assertEqual(expected, mock_callback.call_count)
 
     def test_ensure_one_retry(self):
         mock_callback = mock.Mock(side_effect=IOError)
         self.assertRaises(oslo_messaging.MessageDeliveryFailure,
                           self.connection.ensure, mock_callback,
                           retry=1)
-        self.assertEqual(3, mock_callback.call_count)
+        # TODO(stephenfin): Remove when we drop support for kombu < 5.2.4
+        expected = 2
+        if kombu.VERSION < (5, 2, 4):
+            expected = 3
+        self.assertEqual(expected, mock_callback.call_count)
 
     def test_ensure_no_retry(self):
         mock_callback = mock.Mock(side_effect=IOError)
-        self.assertRaises(oslo_messaging.MessageDeliveryFailure,
-                          self.connection.ensure, mock_callback,
-                          retry=0)
-        self.assertEqual(2, mock_callback.call_count)
+        self.assertRaises(
+            oslo_messaging.MessageDeliveryFailure,
+            self.connection.ensure,
+            mock_callback,
+            retry=0,
+        )
+        # TODO(stephenfin): Remove when we drop support for kombu < 5.2.4
+        expected = 1
+        if kombu.VERSION < (5, 2, 4):
+            expected = 2
+        self.assertEqual(expected, mock_callback.call_count)
 
 
 class ConnectionLockTestCase(test_utils.BaseTestCase):
@@ -1080,3 +1095,82 @@ class ConnectionLockTestCase(test_utils.BaseTestCase):
         t2 = self._thread(lock, 1)
         self.assertAlmostEqual(1, t1(), places=0)
         self.assertAlmostEqual(2, t2(), places=0)
+
+
+class TestPollTimeoutLimit(test_utils.BaseTestCase):
+    def test_poll_timeout_limit(self):
+        transport = oslo_messaging.get_transport(self.conf,
+                                                 'kombu+memory:////')
+        self.addCleanup(transport.cleanup)
+        driver = transport._driver
+        target = oslo_messaging.Target(topic='testtopic')
+        listener = driver.listen(target, None, None)._poll_style_listener
+
+        thread = threading.Thread(target=listener.poll)
+        thread.daemon = True
+        thread.start()
+        time.sleep(amqpdriver.ACK_REQUEUE_EVERY_SECONDS_MAX * 2)
+
+        try:
+            # timeout should not grow past the maximum
+            self.assertEqual(amqpdriver.ACK_REQUEUE_EVERY_SECONDS_MAX,
+                             listener._current_timeout)
+
+        finally:
+            # gracefully stop waiting
+            driver.send(target,
+                        {},
+                        {'tx_id': 'test'})
+            thread.join()
+
+
+class TestMsgIdCache(test_utils.BaseTestCase):
+    @mock.patch('kombu.message.Message.reject')
+    def test_reply_wire_format(self, reject_mock):
+        self.conf.oslo_messaging_rabbit.kombu_compression = None
+
+        transport = oslo_messaging.get_transport(self.conf,
+                                                 'kombu+memory:////')
+        self.addCleanup(transport.cleanup)
+
+        driver = transport._driver
+
+        target = oslo_messaging.Target(topic='testtopic',
+                                       server=None,
+                                       fanout=False)
+
+        listener = driver.listen(target, None, None)._poll_style_listener
+
+        connection, producer = _create_producer(target)
+        self.addCleanup(connection.release)
+
+        msg = {
+            'oslo.version': '2.0',
+            'oslo.message': {}
+        }
+
+        msg['oslo.message'].update({
+            '_msg_id': uuid.uuid4().hex,
+            '_unique_id': uuid.uuid4().hex,
+            '_reply_q': 'reply_' + uuid.uuid4().hex,
+            '_timeout': None,
+        })
+
+        msg['oslo.message'] = jsonutils.dumps(msg['oslo.message'])
+
+        producer.publish(msg)
+
+        received = listener.poll()[0]
+        self.assertIsNotNone(received)
+        self.assertEqual({}, received.message)
+
+        # publish the same message a second time
+        producer.publish(msg)
+
+        received = listener.poll(timeout=1)
+
+        # duplicate message is ignored
+        self.assertEqual(len(received), 0)
+
+        # we should not reject duplicate message
+        reject_mock.assert_not_called()
